/*
 *  This file is part of Subsequence Mathing Framework (SMF) library
 *
 *  SMF library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Skip Graphs library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Skip Graphs library.  If not, see <http://www.gnu.org/licenses/>.
 */
package smf.algorithms;

import java.lang.reflect.InvocationTargetException;
import java.util.*;
import java.util.logging.Level;
import java.util.logging.Logger;
import messif.algorithms.Algorithm;
import messif.objects.LocalAbstractObject;
import messif.objects.util.RankedAbstractObject;
import messif.operations.query.KNNQueryOperation;
import messif.utility.reflection.NoSuchInstantiatorException;
import smf.exceptions.StorageException;
import smf.modules.DistanceIndex;
import smf.modules.SequenceSlicer;
import smf.modules.SequenceStorage;
import smf.modules.slicer.AllPossibleWindowsSlicer;
import smf.sequences.DistanceAllowsNonEquilength;
import smf.sequences.Sequence;
import smf.sequences.SequenceFactory;
import smf.utils.RankedUniqueSubsequenceParentCollection;

/**
 * This algorithm provides functionality for the subsequence matching with
 * variable query length on the original data. Post processing with {@link smf.modules.slicer.AllPossibleWindowsSlicer}.
 * 
 * TODO: allow to change postprocess slicer
 *
 * @param <T> type of internal sequence data (e.g. float[])
 * @author Jakub Valcik, Masaryk University, Brno, Czech Republic,
 * xvalcik@fi.muni.cz
 */
public class VariableResultLengthAlgorithm<T> extends VariableQueryAlgorithm<T> {

    /**
     * Serial class ID for serialization
     */
    private static final long serialVersionUID = 97301L;
    /**
     * Logger instance used in the whole SMF framework
     */
    private static final Logger logger = Logger.getLogger("SMF");
    

    @Algorithm.AlgorithmConstructor(description = "Subsequence matching algorithm with variable result length", arguments = {"sequence object class", "whole sequence storage", "subsequence storage", "data slicer", "query slicer", "window size"})
    public VariableResultLengthAlgorithm(Class<? extends Sequence<T>> objectClass, SequenceStorage<T> sequenceStorage,
            DistanceIndex distanceIndex, SequenceSlicer dataSlicer, SequenceSlicer querySlicer, int windowSize) throws IllegalArgumentException {
        this(objectClass, sequenceStorage, distanceIndex, dataSlicer, querySlicer, windowSize, 100);
    }

    @Algorithm.AlgorithmConstructor(description = "Subsequence matching algorithm with variable query length", arguments = {"sequence object class", "whole sequence storage", "subsequence storage", "data slicer", "query slicer", "window size", "# of candidate slices"})
    public VariableResultLengthAlgorithm(Class<? extends Sequence<T>> objectClass, SequenceStorage<T> sequenceStorage,
            DistanceIndex distanceIndex, SequenceSlicer dataSlicer, SequenceSlicer querySlicer, int windowSize, int kNNSlices) throws IllegalArgumentException {
        super("Subsequence matching algorithm with variable result length", objectClass, sequenceStorage, distanceIndex, dataSlicer, querySlicer, windowSize, kNNSlices);
    }

    /**
     * Performs KNNQueryOperation. It returns the K best matches with respect to
     * the distance function implemented in a given sequence object
     *
     * @param operation
     * @throws StorageException
     */
    @Override
    public void kNNSearch(KNNQueryOperation operation) throws StorageException {
        try {
            operation.setAnswerCollection(new RankedUniqueSubsequenceParentCollection(operation));

            Sequence<T> querySequence = ((Sequence<T>) operation.getQueryObject());
            logger.log(Level.FINER, "Query offset: {0}", querySequence.getOffset());
            boolean allowsNonEquilength = (operation.getQueryObject() instanceof DistanceAllowsNonEquilength);
            boolean setParentSeq = Boolean.valueOf(operation.getParameter(Sequence.PARAM_SET_PARENT_SEQ, String.class, "true"));

            List<LocalAbstractObject> querySubsequenceList = SequenceFactory.create(objectClass.cast(operation.getQueryObject()), querySlicer);
            Map<String, TreeSet<Integer>> sequenceIntervalMap = new HashMap<String, TreeSet<Integer>>();

            // For each query subsequence run a separate query
            for (LocalAbstractObject o : querySubsequenceList) {

                Iterator<RankedAbstractObject> answerIterator = distanceIndex.findNearest(o, kNNSlices);
                Sequence querySubsequence = (Sequence) o;
                logger.log(Level.FINER, "Query slice offset: {0}", querySubsequence.getOffset());

                // For each retrieved subsequence answer
                while (answerIterator.hasNext()) {
                    Sequence<T> subseqAnswer = (Sequence<T>) answerIterator.next().getObject();

                    if (!sequenceIntervalMap.containsKey(subseqAnswer.getOriginalSequenceLocator())) {
                        sequenceIntervalMap.put(subseqAnswer.getOriginalSequenceLocator(), new TreeSet<Integer>());
                    }


                    int sliceOffset = subseqAnswer.getOffset() - 2 * (querySubsequence.getOffset() + querySubsequence.getSequenceLength() / 2);
                    if (sliceOffset < 0) {
                        sliceOffset = 0;
                    }
                    sequenceIntervalMap.get(subseqAnswer.getOriginalSequenceLocator()).add(sliceOffset);
                }
            }

            logger.log(Level.INFO, "Candidate sequences count: {0}", sequenceIntervalMap.size());

            int intervalsCount = 0;
            int dtwCount = 0;

            for (Iterator<String> candidateSequenceIt = sequenceIntervalMap.keySet().iterator(); candidateSequenceIt.hasNext();) {
                String candidateSequenceLocator = candidateSequenceIt.next();
                Sequence<T> candidateSequence = sequenceStorage.getSequence(candidateSequenceLocator);

                Iterator<Integer> offsetIt = sequenceIntervalMap.get(candidateSequenceLocator).iterator();
                int startOffset = offsetIt.next();
                int subseqTo = startOffset + 2 * querySequence.getSequenceLength();
                while (offsetIt.hasNext()) {
                    int offset = offsetIt.next();
                    if (offset > subseqTo) {
                        ++intervalsCount;
                        dtwCount += examineInterval(operation, candidateSequence, startOffset, subseqTo, allowsNonEquilength, setParentSeq);
                        startOffset = offset;
                    }
                    subseqTo = offset + 2 * querySequence.getSequenceLength();
                }
                ++intervalsCount;
                dtwCount += examineInterval(operation, candidateSequence, startOffset, subseqTo, allowsNonEquilength, setParentSeq);
            }
            logger.log(Level.INFO, "Examined intervals count: {0}", intervalsCount);
            logger.log(Level.INFO, "Distance computation count: {0}", dtwCount);
            logger.log(Level.INFO, "Processed kNN query of length {0}, window size = {1}\n\tthe query initiated {2} subsequent queries to the distane index", new Object[]{((Sequence) operation.getQueryObject()).getSequenceLength(), windowSize, querySubsequenceList.size()});
            operation.endOperation();
        } catch (NoSuchInstantiatorException ex) {
            Logger.getLogger(VariableResultLengthAlgorithm.class.getName()).log(Level.SEVERE, null, ex);
        } catch (InvocationTargetException ex) {
            Logger.getLogger(VariableResultLengthAlgorithm.class.getName()).log(Level.SEVERE, null, ex);
        }
    }

    private int examineInterval(KNNQueryOperation operation, Sequence<T> candidateSequence, int offset, int to, boolean allowsNonEquilength, boolean setParentSeq) throws NoSuchInstantiatorException, InvocationTargetException {
        //narezat na vsechny kousicky
        Sequence<T> querySequence = (Sequence<T>) operation.getQueryObject();
        AllPossibleWindowsSlicer allPossibleWinSlicer = new AllPossibleWindowsSlicer(offset, to - offset, querySequence.getSequenceLength() / 2, querySequence.getSequenceLength() * 2);
        List<Sequence<T>> alignedDataSubsequenceSlices = SequenceFactory.create(objectClass.cast(candidateSequence), allPossibleWinSlicer, setParentSeq);
        logger.log(Level.INFO, "Slices count: {0} [{1} - {2}] #{3}#{4}", new Object[]{alignedDataSubsequenceSlices.size(), offset, to, ((LocalAbstractObject)candidateSequence).getLocatorURI(), candidateSequence.getOriginalSequenceLocator()});
        int distanceComputations = 0;
        for (Iterator<Sequence<T>> sliceIt = alignedDataSubsequenceSlices.iterator(); sliceIt.hasNext();) {
            Sequence<T> alignedDataSubsequenceSlice = sliceIt.next();
            if (alignedDataSubsequenceSlice.getSequenceLength() != querySequence.getSequenceLength() && !allowsNonEquilength) {
                continue;
            }
            // Distance between the query sequence and aligned subsequence sliced from the answer subsequence's original sequence
            float dist = ((LocalAbstractObject) alignedDataSubsequenceSlice).getDistance((LocalAbstractObject) querySequence) / alignedDataSubsequenceSlice.getSequenceLength();
            //logger.log(Level.FINE, "Adding sequence to the answer collection...");
            operation.addToAnswer((LocalAbstractObject) alignedDataSubsequenceSlice, dist, null);
            ++distanceComputations;
        }

        return distanceComputations;
    }

    
    /**
     * Properly handles end of the work with storages
     *
     * @throws Throwable
     */
    @SuppressWarnings({"FinalizeDeclaration", "FinalizeCalledExplicitly"})
    @Override
    public void finalize() throws Throwable {
        super.finalize();
    }

    /**
     * Releases resources associated with local and sub storages
     *
     * @throws Throwable
     */
    @Override
    public void destroy() throws Throwable {
        super.destroy();
    }
}
