/*
 *  This file is part of Subsequence Mathing Framework (SMF) library
 *
 *  SMF library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Skip Graphs library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Skip Graphs library.  If not, see <http://www.gnu.org/licenses/>.
 */
package smf.modules.distance;

import java.util.Stack;
import messif.objects.DistanceFunction;
import smf.sequences.DistanceAllowsNonEquilength;
import smf.sequences.Sequence;

/**
 * This method encapsulates an implementation of the DTW distance for sequences.
 * 
 * @param <T> type of the sequence data
 * @author David Novak, Masaryk University, Brno, Czech Republic, david.novak@fi.muni.cz
 * @author Petr Volny, Masaryk University, Brno, Czech Republic, volny.petr@gmail.com
*/
public class DTWSequenceDist<T> implements DistanceFunction<Sequence<T>>,DistanceAllowsNonEquilength {
    
    /**
     * Computes DTW distance between two sequences
     * @param s1 first Sequence object
     * @param s2 second Sequence object
     * @return DTW distance between the first and the second sequence object. If at least one of the given objects is not of a type Sequence than -1 is returned
     */
    @Override
    public float getDistance(Sequence<T> s1, Sequence<T> s2) {
        float res = 0f;
        Stack<Float> w = new Stack<Float>(); // warping path storage
        int n = s1.getSequenceLength(), m = s2.getSequenceLength();
        float d[][] = new float[n][m]; // local distances matrix
        float cm[][] = new float[n][m]; // accum matrix
        for (int i = 0; i < n; i++) {
            for (int j = 0; j < m; j++) {
                d[i][j] = s1.getPiecewiseDist(i, s2, j);
            }
        }
        // create the accum matrix
        cm[0][0] = d[0][0];
        for (int i = 1; i < n; i++) {
            cm[i][0] = d[i][0] + cm[i - 1][0];
        }
        for (int j = 1; j < m; j++) {
            cm[0][j] = d[0][j] + cm[0][j - 1];
        }
        // Compute the matrix values
        for (int i = 1; i < n; i++) {
            for (int j = 1; j < m; j++) {
                // Decide on the path with minimum distance so far
                if (cm[i - 1][j - 1] <= cm[i - 1][j]) {
                    if (cm[i - 1][j - 1] <= cm[i][j - 1]) {
                        cm[i][j] = d[i][j] + cm[i - 1][j - 1];
                    } else {
                        cm[i][j] = d[i][j] + cm[i][j - 1];
                    }
                } else if (cm[i - 1][j] <= cm[i][j - 1]) {
                    cm[i][j] = d[i][j] + cm[i - 1][j];
                } else {
                    cm[i][j] = d[i][j] + cm[i][j - 1];
                }
            }
        }
        int i = n - 1, j = m - 1;
        // find the warping path
        w.push(new Float(cm[i][j]));
        res += cm[i][j];
        do {
            if (i > 0 && j > 0) {
                if (cm[i - 1][j - 1] <= cm[i - 1][j]) {
                    if (cm[i - 1][j - 1] <= cm[i][j - 1]) {
                        i--;
                        j--;
                    } else {
                        j--;
                    }
                } else if (cm[i - 1][j] <= cm[i][j - 1]) {
                    i--;
                } else {
                    j--;
                }
            } else if (i == 0) {
                j--;
            } else {
                i--;
            }
            w.push(new Float(cm[i][j]));
            res += cm[i][j];
        } while (i != 0 || j != 0);
        return res;
    }

    @Override
    public Class<? extends Sequence<T>> getDistanceObjectClass() {
        throw new UnsupportedOperationException("Not supported yet.");
        // TODO: check with Michal
    }

}
