/*
 *  This file is part of Subsequence Mathing Framework (SMF) library
 *
 *  SMF library is free software: you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation, either version 3 of the License, or
 *  (at your option) any later version.
 *
 *  Skip Graphs library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with Skip Graphs library.  If not, see <http://www.gnu.org/licenses/>.
 */
package smf.utils;

import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;

import messif.objects.util.RankedAbstractObject;
import messif.objects.util.RankedSortedCollection;
import messif.operations.RankingQueryOperation;
import smf.sequences.Sequence;

/**
 * This collections maintains only such objects (subsequences) that are unique, which means either 
 *  they are from different parent sequence or they do not overlap.
 * 
 * @author David Novak, Masaryk University, Brno, Czech Republic, david.novak@fi.muni.cz
 * @author Petr Volny, Masaryk University, Brno, Czech Republic, volny.petr@gmail.com
 */
public class RankedUniqueSubsequenceParentCollection extends RankedSortedCollection {

    public RankedUniqueSubsequenceParentCollection() throws IllegalArgumentException {
        super();
    }

    public RankedUniqueSubsequenceParentCollection(RankingQueryOperation operation) {
        super(operation);
    }

    /**
     * Adds an object (sequence) only if this sequence (overlapping) is not in the collection already.
     * @param e RankedAbstractObject to be added to the answer
     * @return true if the RankedAbstractObject was added to the answer
     */
    @Override
    public boolean add(RankedAbstractObject e) {
        Sequence eSeq = (Sequence) e.getObject();
        if (eSeq.getOriginalSequenceLocator() == null) {
            Logger.getLogger(RankedUniqueSubsequenceParentCollection.class.getName()).log(Level.SEVERE, "OriginalSequenceLocator is NULL");
            return false;
        }

        // Iterate through the objects added to the answer so far
        for (Iterator<RankedAbstractObject> it = this.iterator(); it.hasNext();) {            
            RankedAbstractObject itObj = it.next();
            Sequence itObjSeq = (Sequence) itObj.getObject();
            if (itObjSeq.getOriginalSequenceLocator().equals(eSeq.getOriginalSequenceLocator())) {
                // Find out whether the sequences overlaping each other
                int offsetDist = itObjSeq.getOffset() - eSeq.getOffset();
                if (offsetDist >= 0 ? offsetDist < eSeq.getSequenceLength() : - offsetDist < itObjSeq.getSequenceLength()) {
                    // Compare distances
                    if (e.getDistance() < itObj.getDistance()) {
                        it.remove();
                        return super.add(e);
                    } else {
                        return false;
                    }
                }
            }
        }
        return super.add(e);
    }
}
